﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;

internal class FindTextDialog : IDisposable
{
    const string FINDMSGSTRINGW = "commdlg_FindReplace";
    const string FINDMSGSTRING = FINDMSGSTRINGW;

    const int WM_DESTROY = 0x0002;

    const uint FR_DIALOGTERM = 0x00000040;
    const uint FR_FINDNEXT = 0x00000008;

    [DllImport("user32.dll", SetLastError = true, CharSet = CharSet.Auto)]
    static extern uint RegisterWindowMessage(string lpString);

    [DllImport("user32.dll", CharSet = CharSet.Auto)]
    static extern IntPtr SendMessage(IntPtr hWnd, uint Msg, nuint wParam, nint lParam);

    [DllImport("Comdlg32.dll", SetLastError = true, CharSet = CharSet.Auto)]
    static extern IntPtr FindText(ref FINDREPLACEW unnamedParam1);

    [DllImport("user32.dll", CharSet = CharSet.Auto)]
    static extern bool DestroyWindow(IntPtr hWnd);

    uint _findMsgString = 0;
    public bool IsFindMessage(int msg) => msg == (int)_findMsgString;

    FINDREPLACEW _findItem = new FINDREPLACEW();

    IntPtr _pFindWhatBuffer = IntPtr.Zero;
    IntPtr _hWnd = IntPtr.Zero;
    public IntPtr Handle => _hWnd;
    
    IntPtr _hwndParent = IntPtr.Zero;

    public string Text
    {
        get
        {
            if (_pFindWhatBuffer == IntPtr.Zero)
            {
                return "";
            }
            
            return Marshal.PtrToStringAuto(_pFindWhatBuffer) ?? "";
        }
    }

    public bool HasFindNextFlag()
    {
        return (_findItem.Flags & FR_FINDNEXT) == FR_FINDNEXT;
    }

    public bool HasCloseFlag()
    {
        if (HasFindNextFlag())
        {
            return false;
        }
        
        return (_findItem.Flags & FR_DIALOGTERM) == FR_DIALOGTERM;
    }

    public unsafe FindTextDialog(IntPtr owner)
    {
        if (IntPtr.Size == 8)
        {
            Debug.Assert(Marshal.SizeOf(_findItem) == 0x50);
        }

        _hwndParent = owner;

        _pFindWhatBuffer = new IntPtr(Marshal.AllocHGlobal(80));
        Unsafe.InitBlockUnaligned((byte*)_pFindWhatBuffer.ToPointer(), 0, 80);

        if (_findMsgString == 0)
        {
            _findMsgString = RegisterWindowMessage(FINDMSGSTRING);
        }

        _findItem.lStructSize = (uint)Marshal.SizeOf(_findItem);
        _findItem.hwndOwner = _hwndParent;
        _findItem.hInstance = Marshal.GetHINSTANCE(typeof(FindTextDialog).Module);
        _findItem.lpstrFindWhat = _pFindWhatBuffer;
        _findItem.wFindWhatLen = 80;
    }

    public void CreateDialog()
    {
        if (_hWnd != IntPtr.Zero)
        {
            return;
        }
        
        if (_findMsgString != 0)
        {
            _hWnd = FindText(ref _findItem);
        }
    }

    public void Close()
    {
        if (_hWnd != IntPtr.Zero)
        {
            // SendMessage(_hWnd, WM_DESTROY, UIntPtr.Zero, IntPtr.Zero);
            DestroyWindow(_hWnd);
            _hWnd = IntPtr.Zero;
        }
    }

    public void Dispose()
    {
        Close();

        if (_pFindWhatBuffer != IntPtr.Zero)
        {
            Marshal.FreeHGlobal(_pFindWhatBuffer);
            _pFindWhatBuffer = IntPtr.Zero;
        }
    }
}

[StructLayout(LayoutKind.Sequential)]
unsafe struct FINDREPLACEW
{
    public uint lStructSize;        // size of this struct 0x20
    public IntPtr hwndOwner;          // handle to owner's window
    public IntPtr hInstance;          // instance handle of.EXE that
                                      //   contains cust. dlg. template
    public uint Flags;              // one or more of the FR_??
    public IntPtr lpstrFindWhat;      // ptr. to search string
    public char* lpstrReplaceWith;   // ptr. to replace string
    public ushort wFindWhatLen;       // size of find buffer
    public ushort wReplaceWithLen;    // size of replace buffer

    public IntPtr lCustData;          // data passed to hook fn.
    public delegate*<IntPtr, uint, uint, IntPtr, IntPtr> lpfnHook;           // ptr. to hook fn. or NULL
    public char* lpTemplateName;     // custom template name
}

[StructLayout(LayoutKind.Sequential)]

public struct MSG
{
    IntPtr hwnd;
    uint message;
    UIntPtr wParam;
    IntPtr lParam;
    int time;
    POINT pt;
    int lPrivate;
    
    public static MSG From(Message m)
    {
        MSG msg = new MSG();
        msg.hwnd = m.HWnd;
        msg.message = (uint)m.Msg;
        msg.wParam = (UIntPtr)m.WParam;
        msg.lParam = m.LParam;
        msg.time = 0;
        msg.lParam = 0;

        return msg;
    }
}

[StructLayout(LayoutKind.Sequential)]
public struct POINT
{
    public int X;
    public int Y;

    public POINT(int x, int y)
    {
        this.X = x;
        this.Y = y;
    }

    public static implicit operator System.Drawing.Point(POINT p)
    {
        return new System.Drawing.Point(p.X, p.Y);
    }

    public static implicit operator POINT(System.Drawing.Point p)
    {
        return new POINT(p.X, p.Y);
    }

    public override string ToString()
    {
        return $"X: {X}, Y: {Y}";
    }
}
